import BigNumber from 'bignumber.js';
import { ApolloClient, gql, InMemoryCache } from '@apollo/client';
import { DetailedFungibleAssetOnChain } from '@onomy/hub';
import { useAsyncPoll, useAsyncStatuses } from '@onomy/react-utils';
import { useCallback, useMemo, useState } from 'react';
import { useOnomyHub } from '@onomy/react-hub';
import { useOnomyApp } from '../../wallet/providers/OnomyApp';

const ONEX_ORDER_BOOKS_QUERY = gql`
  query orderBooks($base: String!, $quote: String!) {
    orderBooks(where: { base: $base, quote: $quote }) {
      id
      total
      bins
    }
  }
`;

const ONEX_ORDER_BOOK_INCREMENTS_QUERY = gql`
  query bookIncrements($book: String!, $bin: BigDecimal!) {
    bookIncrements(where: { book: $book, bin: $bin }, orderBy: place, orderDirection: desc) {
      id
      place
      orders
      amount
    }
  }
`;

export function useOrderBookBins({
  apolloClient,
  baseAsset,
  quoteAsset,
}: {
  apolloClient: ApolloClient<object>;
  baseAsset: DetailedFungibleAssetOnChain;
  quoteAsset: DetailedFungibleAssetOnChain;
}) {
  const cb = useCallback(async () => {
    const resp = await apolloClient.query({
      query: ONEX_ORDER_BOOKS_QUERY,
      variables: {
        base: baseAsset.id,
        quote: quoteAsset.id,
      },
    });

    const {
      data: { orderBooks: books },
    } = resp;

    const normalized = books.map((book: any) => {
      return {
        ...book,
        bins: book.bins.map((binId: string) => binId.split('-').pop()),
      };
    });

    // TODO: this assumes buy/sell bins always match
    const bins: string[] = (normalized[0]?.bins ?? []).sort((a: string, b: string) => {
      return parseFloat(a) - parseFloat(b);
    });
    return bins;
  }, [apolloClient, baseAsset.id, quoteAsset.id]);

  return useAsyncPoll(cb, [] as Awaited<ReturnType<typeof cb>>);
}

export function useOrderBookIncrements({
  apolloClient,
  baseAsset,
  quoteAsset,
  bin,
  direction,
}: {
  apolloClient: ApolloClient<object>;
  baseAsset: DetailedFungibleAssetOnChain;
  quoteAsset: DetailedFungibleAssetOnChain;
  bin: string;
  direction: 'buy' | 'sell';
}) {
  const { hub } = useOnomyHub();
  const cb = useCallback(async () => {
    if (!bin || !hub) return [];
    const book = `${baseAsset.id}-${quoteAsset.id}-${direction}`;
    const resp = await apolloClient.query({
      query: ONEX_ORDER_BOOK_INCREMENTS_QUERY,
      variables: {
        book,
        bin,
      },
    });

    let {
      data: { bookIncrements = [] },
    } = resp;

    let totalBase = new BigNumber(0);

    try {
      if (direction === 'sell') bookIncrements = bookIncrements.slice().reverse();
    } catch (e) {
      console.error(e);
    }

    const increments = (bookIncrements as any[]).map((raw: any) => {
      // TODO: api doesn't account for potential difference in decimal places
      const bnPlace = new BigNumber(raw.place);
      const bnPricePer = bnPlace.shiftedBy(quoteAsset.meta.decimalPlaces);
      const bnQuoteAmount = bnPlace.times(raw.amount);

      totalBase = totalBase.plus(raw.amount);

      return {
        uid: raw.id as string,
        place: raw.place as string,
        amountRaw: raw.amount as string,
        orders: raw.orders as string[],
        amount: {
          amount: BigInt(raw.amount),
          asset: baseAsset,
        },
        totalAmount: {
          amount: BigInt(totalBase.toFixed(0)),
          asset: baseAsset,
        },
        quoteAmount: {
          amount: BigInt(bnQuoteAmount.toFixed(0)),
          asset: quoteAsset,
        },
        pricePer: {
          amount: BigInt(bnPricePer.toFixed(0)),
          asset: quoteAsset,
        },
      };
    });

    if (direction === 'sell') increments.reverse();

    const detailedIncrements = await Promise.all(
      increments.map(async inc => {
        const [amount, totalAmount, quoteAmount, pricePer] =
          await hub.balances.includeDetailsMultiple([
            inc.amount,
            inc.totalAmount,
            inc.quoteAmount,
            inc.pricePer,
          ]);
        return {
          ...inc,
          percentage: new BigNumber(totalAmount.amount.toString())
            .dividedBy(totalBase)
            .times(100)
            .toNumber(),
          amount,
          totalAmount,
          quoteAmount,
          pricePer,
        };
      })
    );

    return detailedIncrements;
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [hub, apolloClient, baseAsset.id, quoteAsset.id, bin, direction]);

  return useAsyncPoll(cb, [] as Awaited<ReturnType<typeof cb>>);
}

export function useIndexedOrderBookForm({
  baseAsset,
  quoteAsset,
}: {
  baseAsset: DetailedFungibleAssetOnChain;
  quoteAsset: DetailedFungibleAssetOnChain;
}) {
  const { ecosystem } = useOnomyApp();
  const apolloClient = useMemo(() => {
    return new ApolloClient({
      uri: ecosystem.onexGraphqlEndpoint,
      cache: new InMemoryCache(),
    });
  }, [ecosystem.onexGraphqlEndpoint]);

  const [orderBins, orderBinsStatus] = useOrderBookBins({
    apolloClient,
    baseAsset,
    quoteAsset,
  });

  const binOptions = useMemo(
    () =>
      orderBins.map(bin => {
        const val = parseFloat(bin);

        return {
          value: bin,
          label: val > 1 ? val.toLocaleString() : val.toString(),
        };
      }),
    [orderBins]
  );

  const [selectedBin, setSelectedBin] = useState<undefined | string>(undefined);

  useMemo(() => {
    if (binOptions[0] && typeof selectedBin === 'undefined') {
      setSelectedBin(binOptions[0].value);
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [binOptions[0]?.value]);

  const [bookBuy, bookBuyStatus] = useOrderBookIncrements({
    apolloClient,
    baseAsset,
    quoteAsset,
    bin: selectedBin ?? '',
    direction: 'buy',
  });

  const [bookSell, bookSellStatus] = useOrderBookIncrements({
    apolloClient,
    baseAsset,
    quoteAsset,
    bin: selectedBin ?? '',
    direction: 'sell',
  });

  const loadingStatus = useAsyncStatuses([orderBinsStatus, bookBuyStatus, bookSellStatus]);

  const chartData = useMemo(() => {
    const bids = bookBuy.map(item => {
      return {
        side: 'buy',
        price: parseFloat(item.pricePer.amountFormatted),
        amount: parseFloat(item.amount.amountFormatted),
      };
    });
    const asks = bookSell.map(item => {
      return {
        side: 'sell',
        price: parseFloat(item.pricePer.amountFormatted),
        amount: parseFloat(item.amount.amountFormatted),
      };
    });

    return {
      market: `${baseAsset.meta.symbol}/${quoteAsset.meta.symbol}`,
      bids,
      asks,
    };
  }, [bookBuy, bookSell, baseAsset.meta.symbol, quoteAsset.meta.symbol]);

  return {
    loadingStatus,
    orderBinsStatus,
    bookBuyStatus,
    bookSellStatus,

    binOptions,
    selectedBin,
    bookBuy,
    bookSell,
    chartData,

    setSelectedBin,
  };
}
