import React, { createContext, useContext, useMemo } from 'react';
import { useAsyncStatuses } from '@onomy/react-utils';
import { useDetailedBalance, useProposal, useProposalMinDeposit } from '@onomy/react-hub';

import { PlaceholderErrorDisplay } from '../../components/PlaceholderErrorDisplay';
import { calculateVotesChartData } from '../../governanceUtils';

function useProposalDetailState(props: Parameters<typeof useProposal>[0]) {
  const [proposal, proposalStatus] = useProposal(props);

  const votesChartData = useMemo(() => {
    return proposal ? calculateVotesChartData(proposal) : [0, 0, 0];
  }, [proposal]);

  const isVotingPeriod = proposal?.status === 'PROPOSAL_STATUS_VOTING_PERIOD';
  const isDepositPeriod = proposal?.status === 'PROPOSAL_STATUS_DEPOSIT_PERIOD';

  const [minDepositAmount, minDepositAmountStatus] = useProposalMinDeposit(props.chain);

  // TODO: support multiple deposit denoms?
  const depositNeededFixed =
    minDepositAmount.amount - (proposal?.totalDeposits[0]?.amount ?? BigInt(0));
  const [depositNeeded, depositNeededStatus] = useDetailedBalance(
    minDepositAmount.asset,
    depositNeededFixed.toString(10)
  );

  const loadingStatus = useAsyncStatuses([
    proposalStatus,
    minDepositAmountStatus,
    depositNeededStatus,
  ]);

  return {
    chain: props.chain,
    proposalId: props.id,

    proposal,
    votesChartData,
    isVotingPeriod,
    isDepositPeriod,
    minDepositAmount,
    depositNeeded,

    loadingStatus,
  };
}

export type ProposalDetailContextType = ReturnType<typeof useProposalDetailState>;

const ProposalDetailContext = createContext<ProposalDetailContextType>(null as any);

export function useProposalDetail() {
  const state = useContext(ProposalDetailContext);
  if (!state) throw new Error('ProposalDetailContext missing');
  return state;
}

function defaultRenderError(error: any) {
  return <PlaceholderErrorDisplay error={error} />;
}

export function ProposalDetailProvider({
  children,
  renderLoading,
  renderError = defaultRenderError,
  ...props
}: {
  children: JSX.Element | JSX.Element[];
  renderLoading?: () => JSX.Element | JSX.Element[] | null;
  renderError?: typeof defaultRenderError;
} & Parameters<typeof useProposalDetailState>[0]) {
  const state = useProposalDetailState(props);

  if (renderLoading && state.loadingStatus.pending) return renderLoading();
  if (state.loadingStatus.error) return renderError(state.loadingStatus.error);
  if (!state.proposal && renderLoading) return renderLoading();
  return <ProposalDetailContext.Provider value={state}>{children}</ProposalDetailContext.Provider>;
}
