import React, { createContext, useCallback, useContext, useState } from 'react';
import { useDelegateAssetForm } from '@onomy/react-hub';

import { PlaceholderErrorDisplay } from '../../components/PlaceholderErrorDisplay';
import { PlaceholderLoadingIndicator } from '../../components/PlaceholderLoadingIndicator';

type DelegateAssetProps = Parameters<typeof useDelegateAssetForm>[0];

function useDelegateAssetState(props: DelegateAssetProps) {
  const { undelegate: undelegateProp } = props ?? {};
  const [undelegate, setUndelegate] = useState(undelegateProp);

  const toggleUndelegate = useCallback(() => setUndelegate(u => !u), []);

  const form = useDelegateAssetForm({ ...props, undelegate });
  return {
    ...form,

    undelegate,

    setUndelegate,
    toggleUndelegate,
  };
}

export type DelegateAssetContextType = ReturnType<typeof useDelegateAssetState>;

const DelegateAssetContext = createContext<DelegateAssetContextType>(null as any);

export function useDelegateAsset() {
  const state = useContext(DelegateAssetContext);
  if (!state) throw new Error('DelegateAssetContext missing');
  return state;
}

// eslint-disable-next-line @typescript-eslint/naming-convention
function DEFAULT_LOADING_RENDER_FN() {
  return <PlaceholderLoadingIndicator text="Loading transfer form" />;
}
// eslint-disable-next-line @typescript-eslint/naming-convention
function DEFAULT_ERROR_RENDER_FN(error: any) {
  return <PlaceholderErrorDisplay error={error} />;
}

export function DelegateAssetProvider({
  children,
  renderLoading = DEFAULT_LOADING_RENDER_FN,
  renderError = DEFAULT_ERROR_RENDER_FN,
  ...props
}: {
  children: JSX.Element | JSX.Element[];
  renderLoading?: typeof DEFAULT_LOADING_RENDER_FN;
  renderError?: typeof DEFAULT_ERROR_RENDER_FN;
} & DelegateAssetProps) {
  const state = useDelegateAssetState(props);

  if (state.loadingStatus.pending || !state.delegation) return renderLoading();
  if (state.loadingStatus.error) return renderError(state.loadingStatus.error);
  return <DelegateAssetContext.Provider value={state}>{children}</DelegateAssetContext.Provider>;
}
